/************************************************************************
 *   MODULE  : MEM_ALLC.C
 *
 *   PURPOSE : allocates and locks down memory used during product
 *             caclulations
 *
 *   AUTHOR  : Ron Hawkins & Craig Woodfolk, SAIC (original)
 *             Kevin Russo, SAIC (1)
 *             Melissa Mathews, SAIC (2)
 *
 *   Change     Date          Reason
 *   ------     ----          ------
 *   Original   25 Sept 92    n/a
 *   1          20 Oct 92     overall program update
 *   2          17 Jun 93     converted XVT 2.0 calls to XVT 3.1
 *
 ************************************************************************/

/****************
 * header files *
 ****************/

#include "los.h"

/************************************************************************
 *   FUNCTION     : mem_alloc
 *
 *   PURPOSE      : general memory allocation function
 *
 *   INPUTS       : size    - how many butes to allocate
 *                  handle  - handle to the memory
 *                  pointer - to the memory location
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : make_radials in module MAKE_RDL.C
 *
 ************************************************************************/

#ifdef PROTO
void mem_alloc(long size, GHANDLE *handle, void *pHMEM pointer)
#else
void mem_alloc(size, handle, pointer)
long    size;
GHANDLE *handle;
void    *pHMEM pointer;
#endif

{
   char pHMEM temp;          /* used to set memory to all zeros       */
   long i;                   /* loop counter                          */

   *handle  =  xvt_gmem_alloc(size);
   *pointer = (void pHMEM) xvt_gmem_lock(*handle);
   if (*pointer == NULL)
      xvt_dm_post_fatal_exit("ERROR:  Memory allocation in function mem_alloc() failed");
   temp = (char pHMEM) *pointer;
   for (i=0; i<size; i++)
      temp[i] = 0;
   
}
