/*****************************************************************************/
/* Copyright (c) 1996-1997 Vectaport, Inc.                                   */
/*                                                                           */
/* Permission to use, copy, modify, distribute, and sell this software and   */
/* its documentation for any purpose is hereby granted without fee, provided */
/* that the above copyright notice appear in all copies and that both that   */
/* copyright notice and this permission notice appear in supporting          */
/* documentation, and that the names of the copyright holders not be used in */
/* advertising or publicity pertaining to distribution of the software       */
/* without specific, written prior permission.  The copyright holders make   */
/* no representations about the suitability of this software for any purpose.*/
/* It is provided "as is" without express or implied warranty.               */
/*                                                                           */
/* THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS         */
/* SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.*/
/* IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,        */
/* INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING     */
/* FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/* NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION      */
/* WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.                             */
/*                                                                           */
/*****************************************************************************/

/*
 * demviewer main program.
 */

#include <MapUnidraw/mapcmds.h>
#include <DemUnidraw/demcatalog.h>
#include <DemUnidraw/demcreator.h>
#include <DemUnidraw/demcmds.h>
#include <DemUnidraw/demeditor.h>
#include <DemUnidraw/demhandler.h>

#include <OverlayUnidraw/ovdoer.h>
#include <OverlayUnidraw/ovunidraw.h>

#include <InterViews/world.h>

#ifdef HAVE_ACE
#include <Dispatch/ace_dispatcher.h>
#include <ace/Acceptor.h>
#include <ace/SOCK_Acceptor.h>
#include <ace/Synch.h>
#include <ace/Singleton.h>
#include <OverlayUnidraw/aceimport.h>
#include <ComUnidraw/comterp-acehandler.h>
#endif

#include <stream.h>

DemEditor* main_editor;

// ----------------------------------------
#ifdef HAVE_ACE

// Our Reactor Singleton.
typedef ACE_Singleton<ACE_Reactor, ACE_Null_Mutex> 
	REACTOR;

// Our ACE_Test_and_Set Singleton.
typedef ACE_Singleton<ACE_Test_and_Set <ACE_Null_Mutex, sig_atomic_t>, ACE_Null_Mutex> 
	QUIT_HANDLER;

// ----------------------------------------

class Camera_Handler : public ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_NULL_SYNCH>
  // = TITLE
  //     Receive client message from the remote clients.
  // 
  // = DESCRIPTION
  //     This class demonstrates how to receive messages from remote
  //     clients using the notification mechanisms in the
  //     <ACE_Reactor>.  In addition, it also illustrates how to
  //     utilize the <ACE_Reactor> timer mechanisms, as well.
{
public:
  // = Initialization and termination methods.
  Camera_Handler (void);

  virtual void destroy (void);
  // Ensure dynamic allocation.

  // = Hooks for opening and closing handlers.
  virtual int open (void *);
  virtual int close (u_long);

protected:
  // = Demultiplexing hooks.
  virtual int handle_input (ACE_HANDLE);
  virtual int handle_timeout (const ACE_Time_Value &tv, 
			      const void *arg); 

private:
  char peer_name_[MAXHOSTNAMELEN + 1];
  // Host we are connected to.
};

// Specialize a Camera Acceptor.
typedef ACE_Acceptor <Camera_Handler, ACE_SOCK_ACCEPTOR> 
	Camera_Acceptor;

// Default constructor.

Camera_Handler::Camera_Handler (void)
{
}

void
Camera_Handler::destroy (void)
{
  REACTOR::instance ()->cancel_timer (this);
  this->peer ().close ();
}

int
Camera_Handler::handle_timeout (const ACE_Time_Value &,
				 const void *arg)
{
  ACE_ASSERT (arg == this);
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) handling timeout from this = %u\n", this));
  return 0;
}

// Perform the camera position record receive. 

int
Camera_Handler::handle_input (ACE_HANDLE)
{
  // Perform two recv's to emulate record-oriented semantics.  Note
  // that this code is not entirely portable since it relies on the
  // fact that sizeof (ssize_t) is the same on both the sender and
  // receiver side.  To correctly handle this is painful, and we leave
  // it as an exercise for the reader ;-).

  //  ssize_t len;
  //  ssize_t n = this->peer ().recv ((void *) &len, sizeof len); 
  CameraPositionRecord pos;
  ssize_t n = this->peer ().recv((void *) &pos, sizeof(pos));

  switch (n)
    {
    case -1:
      ACE_ERROR_RETURN ((LM_ERROR, "(%P|%t) %p at host %s\n",
			"client logger", this->peer_name_), -1);      
      /* NOTREACHED */
    case 0:
      cerr << this->peer_name_ << " disconnected from camera port\n";
      return -1;
    case sizeof (pos):
      {
	CameraCmd* camcmd = new CameraCmd(pos.px(), pos.py(),
					  pos.fx(), pos.fy());
	camcmd->SetEditor(main_editor);
	CommandPusher* pusher = new CommandPusher(camcmd);
	pusher->Push();
        unidraw->Update();
	break;
      }
    default:
      ACE_ERROR_RETURN ((LM_ERROR, "(%P|%t) %p at host %s\n",
			"client logger", this->peer_name_), -1);
      /* NOTREACHED */
    }

  return 0;
}

int
Camera_Handler::open (void *)
{
  ACE_INET_Addr addr;
  
  if (this->peer ().get_remote_addr (addr) == -1)
    return -1;
  else
    {
      ACE_OS::strncpy (this->peer_name_, 
		       addr.get_host_name (), 
		       MAXHOSTNAMELEN + 1);

      if (REACTOR::instance ()->register_handler 
	  (this, ACE_Event_Handler::READ_MASK) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, 
			   "(%P|%t) can't register with reactor\n"), -1);
#if 0
      else if (REACTOR::instance ()->schedule_timer
	  (this, 
	  (const void *) this, 
	   ACE_Time_Value (2), 
	   ACE_Time_Value (2)) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, 
			   "can'(%P|%t) t register with reactor\n"), -1);
#endif
      else
        cerr << this->peer_name_ << " connected to camera port\n";

      return 0;
    }
}

// Perform termination activities when deregistered from the
// ACE_Reactor.

int
Camera_Handler::close (u_long)
{
  this->destroy ();
  return 0;
}

#endif /* HAVE_ACE */

/*****************************************************************************/

static PropertyData properties[] = {
    { "*DemEditor*name", "3D Map Viewer" },
    { "*DemEditor*iconName", "demviewer" },
    { "*domain",  "DEM" },
    { "*TextEditor*rows", "10" },
    { "*TextEditor*columns", "40" },
    { "*TextEditor*FileChooser*rows", "10" },
    { "*initialbrush",  "2" },
    { "*initialfgcolor","1" },
    { "*initialbgcolor","10" },
    { "*initialfont",   "4" },
    { "*initialpattern","1" },
    { "*initialarrow", "none" },
    { "*pagewidth", "8.5" },
    { "*pageheight", "11" },
    { "*gridxincr", "8" },
    { "*gridyincr", "8" },
    { "*font1", "-*-courier-medium-r-normal-*-8-*-*-*-*-*-*-* Courier 8" },
    { "*font2", "-*-courier-medium-r-normal-*-10-*-*-*-*-*-*-* Courier 10" },
    { "*font3", "-*-courier-bold-r-normal-*-12-*-*-*-*-*-*-* Courier-Bold 12" },
    { "*font4",
	"-*-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-* Helvetica 12"
    },
    { "*font5",
	"-*-helvetica-medium-r-normal-*-14-*-*-*-*-*-*-* Helvetica 14"
    },
    { "*font6",
	"-*-helvetica-bold-r-normal-*-14-*-*-*-*-*-*-* Helvetica-Bold 14"
    },
    { "*font7",
	"-*-helvetica-medium-o-normal-*-14-*-*-*-*-*-*-* Helvetica-Oblique 14"
    },
    { "*font8",
	"-*-times-medium-r-normal-*-12-*-*-*-*-*-*-*  Times-Roman 12"
    },
    { "*font9", "-*-times-medium-r-normal-*-14-*-*-*-*-*-*-* Times-Roman 14" } ,
    { "*font10", "-*-times-bold-r-normal-*-14-*-*-*-*-*-*-*  Times-Bold 14" },
    { "*font11",
	"-*-times-medium-i-normal-*-14-*-*-*-*-*-*-* Times-Italic 14"
    },
    { "*brush1",	"none" },
    { "*brush2",	"ffff 0" },
    { "*brush3",	"ffff 1" },
    { "*brush4",	"ffff 2" },
    { "*brush5",	"ffff 3" },
    { "*brush6",	"fff0 0" },
    { "*brush7",	"fff0 1" },
    { "*brush8",	"fff0 2" },
    { "*brush9",	"fff0 3" },
    { "*pattern1",	"none" },
    { "*pattern2",	"0.0" },
    { "*pattern3",	"1.0" },
    { "*pattern4",	"0.75" },
    { "*pattern5",	"0.5" },
    { "*pattern6",	"0.25" },
    { "*pattern7",	"1248" },
    { "*pattern8",	"8421" },
    { "*pattern9",	"f000" },
    { "*pattern10",	"8888" },
    { "*pattern11",	"f888" },
    { "*pattern12",	"8525" },
    { "*pattern13",	"cc33" },
    { "*pattern14",	"7bed" },
    { "*fgcolor1",	"Black" },
    { "*fgcolor2",	"Brown 42240 10752 10752" },
    { "*fgcolor3",	"Red" },
    { "*fgcolor4",	"Orange" },
    { "*fgcolor5",	"Yellow" },
    { "*fgcolor6",	"Green" },
    { "*fgcolor7",	"Blue" },
    { "*fgcolor8",	"Indigo 48896 0 65280" },
    { "*fgcolor9",	"Violet 20224 12032 20224" },
    { "*fgcolor10",	"White" },
    { "*fgcolor11",	"LtGray 50000 50000 50000" },
    { "*fgcolor12",	"DkGray 33000 33000 33000" },
    { "*bgcolor1",	"Black" },
    { "*bgcolor2",	"Brown 42240 10752 10752" },
    { "*bgcolor3",	"Red" },
    { "*bgcolor4",	"Orange" },
    { "*bgcolor5",	"Yellow" },
    { "*bgcolor6",	"Green" },
    { "*bgcolor7",	"Blue" },
    { "*bgcolor8",	"Indigo 48896 0 65280" },
    { "*bgcolor9",	"Violet 20224 12032 20224" },
    { "*bgcolor10",	"White" },
    { "*bgcolor11",	"LtGray 50000 50000 50000" },
    { "*bgcolor12",	"DkGray 33000 33000 33000" },
    { "*history",	"20" },
    { "*dem",           "" },
    { "*mapf",          "" },
    { "*texture",       "" },
    { "*nofp",       "true" },
    { "*url",           "" },
    { "*lod",       "false" },
#ifdef HAVE_ACE
    { "*port",          "20012" },
    { "*import",        "20013" },
    { "*queryport",     "20014" },
    { "*cameraport",    "20015" },
#endif
    { nil }
};

static OptionDesc options[] = {
#ifdef HAVE_ACE
    { "-nofp", "*nofp", OptionValueImplicit, "false" },
    { "-fp", "*nofp", OptionValueImplicit, "true" },
    { "-lod", "*lod", OptionValueImplicit, "true" },
    { "-port", "*port", OptionValueNext },
    { "-import", "*import", OptionValueNext },
    { "-queryport", "*queryport", OptionValueNext },
    { "-cameraport", "*cameraport", OptionValueNext },
#endif
    { "-dem", "*dem", OptionValueNext },
    { "-mapf", "*mapf", OptionValueNext },
    { "-texture", "*texture", OptionValueNext },
    { "-url", "*url", OptionValueNext },
    { nil }
};

/*****************************************************************************/

int main (int argc, char** argv) {
#ifdef HAVE_ACE
    Dispatcher::instance(new AceDispatcher(REACTOR::instance()));
#endif

    DemCreator creator;
    DemCatalog* catalog = new DemCatalog("demviewer", &creator);
    OverlayUnidraw* unidraw = new OverlayUnidraw(
        catalog, argc, argv, options, properties
    );

#ifdef HAVE_ACE
    // Acceptor factory.
    Camera_Acceptor* peer_acceptor = new Camera_Acceptor();

    const char* camportstr = catalog->GetAttribute("cameraport");
    int camportnum = atoi(camportstr);
    if (peer_acceptor->open 
	(ACE_INET_Addr (camportnum)) == -1)
      ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "open"), -1);

    else if (REACTOR::instance ()->register_handler 
	     (peer_acceptor, ACE_Event_Handler::READ_MASK) == -1)
      ACE_ERROR_RETURN ((LM_ERROR, 
			 "registering service with ACE_Reactor\n"), -1);
    else
      cerr << "accepting camera port (" << camportnum << ") connections\n";

    UnidrawImportAcceptor* import_acceptor = new UnidrawImportAcceptor();

    const char* importstr = catalog->GetAttribute("import");
    int importnum = atoi(importstr);
    if (import_acceptor->open 
	(ACE_INET_Addr (importnum)) == -1)
        cerr << "demviewer:  unable to open import port " << importnum << "\n";

    else if (IMPORT_REACTOR::instance ()->register_handler 
	     (import_acceptor, ACE_Event_Handler::READ_MASK) == -1)
        cerr << "demviewer:  unable to register UnidrawImportAcceptor with ACE reactor\n";
    else
      cerr << "accepting import port (" << importnum << ") connections\n";

    DemAcceptor* vtk_acceptor = new DemAcceptor();
    const char* qportstr = catalog->GetAttribute("queryport");
    int qportnum = atoi(qportstr);
    if (vtk_acceptor->open(ACE_INET_Addr(qportnum)) == -1)
      cerr << "demviewer: unable to open query port " << qportnum << "\n";
    else if (REACTOR::instance()->register_handler
	     (vtk_acceptor, ACE_Event_Handler::READ_MASK) == -1)
      cerr << "demviewer: unable to register DemAcceptor with ACE reactor\n";
    else
      cerr << "accepting query port (" << qportnum << ") connections\n";

   // Acceptor factory.
    UnidrawComterpAcceptor* uct_acceptor = new UnidrawComterpAcceptor();

    //const char* uctportstr = catalog->GetAttribute("port");
    const char* uctportstr = catalog->GetAttribute("port");
    int uctportnum = atoi(uctportstr);
    if (uct_acceptor->open 
	(ACE_INET_Addr (uctportnum)) == -1)
        cerr << "demviewer:  unable to open comterp port " << uctportnum << "\n";

    else if (COMTERP_REACTOR::instance ()->register_handler 
	     (uct_acceptor, ACE_Event_Handler::READ_MASK) == -1)
        cerr << "demviewer:  unable to register ComterpAcceptor with ACE reactor\n";
    else
      cerr << "accepting comdraw port (" << uctportnum << ") connections\n";

    // Register QUIT_HANDLER to receive SIGINT commands.  When received,
    // QUIT_HANDLER becomes "set" and thus, the event loop below will
    // exit.
    if (REACTOR::instance ()->register_handler 
	     (SIGINT, QUIT_HANDLER::instance ()) == -1)
      ACE_ERROR_RETURN ((LM_ERROR, 
			 "registering service with ACE_Reactor\n"), -1);

#endif
    int exit_status = 0;

    if (argc > 2) {
	cerr << "Usage: demviewer [-port n] [-import n] [-queryport n] [-cameraport n] [-dem file] [-mapf file] [-texture file] [file]" << "\n";
	exit_status = 1;

    } else {
	const char* initial_file = (argc == 2) ? argv[1] : nil;
	DemEditor* ed = new DemEditor(initial_file);
	ed->VehicleName("demviewer_vehicle");
	main_editor = ed;

	unidraw->Open(ed);
	unidraw->Run();
    }

    delete unidraw;
    return exit_status;
}

